/*
  Copyright (c) 2009 nosuke <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <pthread.h>
#include <libspe2.h>

typedef struct spethread_t {
    int id;
    spe_context_ptr_t context;
    pthread_t pthread;
} spethread_t;

static void *
spethread_fn(void *arg)
{
    spethread_t *spethread = (spethread_t *)arg;
    spe_context_ptr_t context = spethread->context;
    unsigned int entry = SPE_DEFAULT_ENTRY;
    spe_stop_info_t stop_info;
    spe_context_run(context, &entry, 0, (void *)(unsigned long)spethread->id,
                    NULL, &stop_info);
    pthread_exit(NULL);
}


#define TAGID 3

int
main(int argc, char *argv[])
{
    int i;
    spe_program_handle_t *program;
    spethread_t spethreads;
    int flags = SPE_MAP_PS;

    unsigned mbox_data;
    unsigned tagstat;

    // DMAХåե16Byte˥饤󤷤֤ǳ
    volatile unsigned dma_buf[4] __attribute__((aligned(16)));

    program = spe_image_open("dma_libspe2_spe");

    spethreads.id = 0;
    spethreads.context = spe_context_create(flags, NULL);
    spe_program_load(spethreads.context, program);

    printf("start SPE program\n");
    pthread_create(&spethreads.pthread, NULL, &spethread_fn,
                   &spethreads);

    // DMAХåեν
    for (i = 0; i < sizeof(dma_buf) / sizeof(dma_buf[0]); i++)
        dma_buf[i] = 0;

    // DMAХåեΥɥ쥹᡼ܥåǼ
    while (spe_out_mbox_read(spethreads.context, &mbox_data, 1) < 1)
        ;
    
    // LSΥХåեdma_buf16Byte񤭹ޤ׵ȯ
    spe_mfcio_put(spethreads.context, mbox_data, (void *)dma_buf, 16, 
                  TAGID, 0, 0);
    // žλޤԤ
    spe_mfcio_tag_status_read(spethreads.context, 1 << TAGID,
                              SPE_TAG_ALL, &tagstat);

    printf("PPE: dma_buf[0] = 0x%08x, tagstat = 0x%08x\n", dma_buf[0], tagstat);
    dma_buf[0] = 0xcbea0123;

    // dma_bufLSΥХåե16Byteɤ߽Ф׵ȯ
    spe_mfcio_get(spethreads.context, mbox_data, (void *)dma_buf, 16,
                  TAGID, 0, 0);
    // žλޤԤ
    spe_mfcio_tag_status_read(spethreads.context, 1 << TAGID,
                              SPE_TAG_ALL, &tagstat);

    while (spe_out_mbox_read(spethreads.context, &mbox_data, 1) < 1)
        ;

    pthread_join(spethreads.pthread, NULL);
    spe_context_destroy(spethreads.context);

    spe_image_close(program);

    printf("done\n");

    return 0;
}
